import logging
import warnings
from pathlib import Path

import bottleneck as bn
import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.nddata import CCDData, NoOverlapError
from astropy.table import Column, vstack
from astropy.time import Time
from astropy.utils.exceptions import AstropyUserWarning
from astropy.wcs import FITSFixedWarning
from ccdproc import ImageFileCollection
from photutils.aperture import CircularAnnulus, CircularAperture, aperture_photometry
from photutils.centroids import centroid_sources
from pydantic import BaseModel, validate_call
from scipy.spatial.distance import cdist

from stellarphot import PhotometryData, SourceListData
from stellarphot.settings import (
    Camera,
    Observatory,
    PhotometryApertures,
    PhotometryOptionalSettings,
    PhotometrySettings,
)

from .source_detection import compute_fwhm

__all__ = [
    "AperturePhotometry",
    "single_image_photometry",
    "multi_image_photometry",
    "faster_sigma_clip_stats",
    "find_too_close",
    "clipped_sky_per_pix_stats",
    "calculate_noise",
]

# Allowed FITS header keywords for exposure values
EXPOSURE_KEYWORDS = ["EXPOSURE", "EXPTIME", "TELAPSE", "ELAPTIME", "ONTIME", "LIVETIME"]


class AperturePhotometry(BaseModel):
    """Class to perform aperture photometry on one or more images"""

    settings: PhotometrySettings

    @validate_call
    def __call__(
        self,
        file_or_directory: str | Path,
        **kwargs,
    ) -> PhotometryData:
        """
        Perform aperture photometry on a single image or a directory of images.

        Parameters
        ----------
        file_or_directory : str or Path
            The file or directory on which to perform aperture photometry.

        logline : str, optional (Default: "single_image_photometry:")
            String to prepend to all log messages, *only used for single image
            photometry*.

        reject_unmatched : bool, optional (Default: True)
            If ``True``, any sources that are not detected on all the images are
            rejected.  If you are interested in a source that can intermittently
            fall below your detection limits, we suggest setting this to ``False``
            so that all sources detected on each image are reported. *Only used for
            multi-image photometry*.

        object_of_interest : str, optional (Default: None)
            Name of the object of interest. The only files on which photometry
            will be done are those whose header contains the keyword ``OBJECT``
            whose value is ``object_of_interest``. *Only used for multi-image
            photometry* to select which files to perform photometry on.

        Returns
        -------
        photom_data : `stellarphot.PhotometryData`
            Photometry data for all the sources on which aperture photometry was
            performed in all the images. This may be a subset of the sources in
            the source list if locations were too close to the edge of any one
            image or to each other for successful aperture photometry.
        """
        # Make sure we have a Path object
        path = Path(file_or_directory)
        if path.is_dir():
            photom_data = multi_image_photometry(path, self.settings, **kwargs)
        elif path.is_file():
            image = CCDData.read(path)
            photom_data = single_image_photometry(
                image, self.settings, fname=str(path), **kwargs
            )
        else:
            raise ValueError(
                f"file_or_directory '{path}' is not a valid file or directory."
            )

        return photom_data


def single_image_photometry(
    ccd_image,
    photometry_settings,
    fname=None,
    logline="single_image_photometry:",
):
    """
    Perform aperture photometry on a single image, with an options for estimating
    the local background from sigma clipped stats of the counts in an annulus around
    the aperture.  If the sky positions of the sources are in the sourcelist and not
    the image positions (x,y), then the function will compute the image positions from
    the sky positions using the WCS of the image.  It assumes the input sky positions
    are in decimal degree units and in the ICRS frame.

    Parameters
    ----------
    ccd_image : `astropy.nddata.CCDData`
        Image on which to perform aperture photometry.  It's headers must contain
        DATE-OBS, FILTER, and exposure time in units of seconds (identified by one
        of the following keywords: EXPOSURE, EXPTIME, TELAPSE, ELAPTIME, ONTIME,
        or LIVETIME). If AIRMASS is available it will be added to `phot_table`.
        This image  must also have a WCS header associated with it if you want to
        use sky positions as inputs.

    photometry_settings : `stellarphot.settings.PhotometrySettings`
        Photometry settings to use for the photometry.  This includes the camera,
        observatory, the aperture and annulus radii to use for the photometry, a map
        of passbands from your filters to AAVSO filter names, and options for the
        photometry. See `stellarphot.settings.PhotometrySettings` for more information.

    fname : str, optional (Default: None)
        Name of the image file on which photometry is being performed, used for
        logging and tracking purposes, the file is not accessed or modified.

    logline : str, optional (Default: "single_image_photometry:")
        String to prepend to all log messages.

    Returns
    -------

    phot_table : `stellarphot.PhotometryData`
        Photometry data for all the locations in which aperture photometry was
        performed.  This may be a subset of the sources in the sourcelist if
        locations were too close to the edge of the image or to each other for
        successful aperture photometry.  If pixel (x/y) positions were used for
        the photometry, but a valid WCS header was not available for `ccd_image`,
        the output 'ra', 'dec', and 'bjd' columns will have np.nan values

    dropped_sources : list
        This of the star_ids of the sources that fell outside the image or were
        too close together and did not have photometry performed.

    Notes
    -----
    The default behavior (set by `use_coordinates`="pixel") for determining
    the placement of apertures is to use the x/y positions in the `sourcelist`.
    This is because in most cases, ra/dec positions are derived from the x/y
    positions and the WCS of the `ccd_image`.  This default reduces unnecessary
    computation and works if the WCS is not very accurate.

    When attempting to process multiple images of the same field, it makes sense
    to use the ra/dec positions in the sourcelist and the WCS of the `ccd_image`
    to compute the x/y positions on each image individually. In this scenario,
    the `use_coordinates` parameter should be set to "sky".
    """

    sourcelist = SourceListData.read(
        photometry_settings.source_location_settings.source_list_file
    )
    camera = photometry_settings.camera
    observatory = photometry_settings.observatory
    photometry_apertures = photometry_settings.photometry_apertures
    photometry_options = photometry_settings.photometry_optional_settings
    logging_options = photometry_settings.logging_settings
    passband_map = photometry_settings.passband_map
    use_coordinates = photometry_settings.source_location_settings.use_coordinates

    # Check that the input parameters are valid
    if not isinstance(ccd_image, CCDData):
        raise TypeError(
            "ccd_image must be a CCDData object, but it is " f"'{type(ccd_image)}'."
        )
    if not isinstance(sourcelist, SourceListData):
        raise TypeError(
            "sourcelist must be a SourceListData object, but it is "
            f"'{type(sourcelist)}'."
        )
    if not isinstance(camera, Camera):
        raise TypeError(f"camera must be a Camera object, but it is '{type(camera)}'.")
    if not isinstance(observatory, Observatory):
        raise TypeError(
            "observatory_location must be a EarthLocation object, but it "
            f"is '{type(observatory)}'."
        )

    if not isinstance(photometry_apertures, PhotometryApertures):
        raise TypeError(
            "photometry_apertures must be a PhotometryApertures object, but it is "
            f"'{type(photometry_apertures)}'."
        )

    if not isinstance(photometry_options, PhotometryOptionalSettings):
        raise TypeError(
            "photometry_options must be a PhotometryOptions object, but it is "
            f"'{type(photometry_options)}'."
        )

    # Set up logging
    logfile = logging_options.logfile
    console_log = logging_options.console_log

    # If we have no handlers, set up logging
    logger = logging.getLogger("single_image_photometry")

    # Use this to catch if logging was already set up by a call
    # from multi_image_photometry before creating new log file handler
    # (Warning: This actually just catches if the logger has any handlers.
    # Probably not the best way to do this.)
    fh_exists = False  # Default to filehandler not existing
    if logger.hasHandlers() is False:
        logger.setLevel(logging.INFO)
        # Set up logging to a file (in addition to any logging below)
        if logfile is not None:
            # by default this appends to existing logfile
            fh = logging.FileHandler(logfile)
            log_format = logging.Formatter("%(levelname)s - %(message)s")
            fh.setFormatter(log_format)
            fh.setLevel(logging.INFO)
            logger.addHandler(fh)
            fh_exists = True

        # Set up logging to console if requested
        if console_log:
            ch = logging.StreamHandler()
        else:  # otherwise effectively suppress output
            ch = logging.NullHandler()
        console_format = logging.Formatter("%(message)s")
        ch.setFormatter(console_format)
        ch.setLevel(logging.INFO)
        logger.addHandler(ch)

    #
    # Check CCDData headers before proceeding
    #
    # Search for exposure keyword in header and set exposure if found
    matched_kw = None
    for kw in EXPOSURE_KEYWORDS:
        if kw in ccd_image.header:
            matched_kw = kw
            break

    if matched_kw is None:
        logger.warning(
            f"{logline} None of the accepted exposure keywords "
            f"({format(', '.join(EXPOSURE_KEYWORDS))}) found in the "
            "header ... SKIPPING THIS IMAGE!"
        )
        return None, None

    exposure = ccd_image.header[matched_kw]

    # Search for other keywords that are required
    try:
        date_obs = ccd_image.header["DATE-OBS"]
    except KeyError:
        logger.warning(
            f"{logline} 'DATE-OBS' not found in CCD image header "
            "... SKIPPING THIS IMAGE!"
        )
        return None, None
    try:
        filter = ccd_image.header["FILTER"]
    except KeyError:
        logger.warning(
            f"{logline} 'FILTER' not found in CCD image header ... "
            "SKIPPING THIS IMAGE!"
        )
        return None, None

    # Set high pixels to NaN (make sure ccd_image.data is a float array first)
    ccd_image.data = ccd_image.data.astype(float)
    ccd_image.data[ccd_image.data > camera.max_data_value.value] = np.nan

    # Extract necessary values from sourcelist structure
    star_ids = sourcelist["star_id"].value
    xs = sourcelist["xcenter"].value
    ys = sourcelist["ycenter"].value
    ra = sourcelist["ra"].value
    dec = sourcelist["dec"].value
    src_cnt = len(sourcelist)

    # If RA/Dec are available attempt to use them to determine the source positions
    if use_coordinates == "sky" and sourcelist.has_ra_dec:
        try:
            imgpos = ccd_image.wcs.world_to_pixel(
                SkyCoord(ra, dec, unit=u.deg, frame="icrs")
            )
            xs, ys = imgpos[0], imgpos[1]
        except AttributeError:
            # No WCS, skip this image
            msg = f"{logline} ccd_image must have a valid WCS to use RA/Dec!"
            logger.warning(msg)
            return None, None
    elif use_coordinates == "sky" and not sourcelist.has_ra_dec:
        raise ValueError(
            "use_coordinates='sky' but sourcelist does not have" "RA/Dec coordinates!"
        )

    # Reject sources that are within an aperture diameter of each other.
    dropped_sources = []
    try:
        too_close = find_too_close(
            sourcelist,
            photometry_apertures.radius,
            pixel_scale=camera.pixel_scale.value,
        )
    except Exception as err:
        # Any failure here is BAD, so raise an error
        raise RuntimeError(
            f"Call to find_too_close() returned {type(err).__name__}: {str(err)}"
        ) from err
    too_close_cnt = np.sum(too_close)
    non_overlap = ~too_close
    msg = (
        f"{logline} {too_close_cnt} of {src_cnt} sources within 2 aperture radii of "
        "nearest neighbor"
    )

    if photometry_options.reject_too_close:
        # Track dropped sources due to being too close together
        dropped_sources.extend(star_ids[too_close].tolist())
        # Remove sources too close together
        star_ids = star_ids[non_overlap]
        xs = xs[non_overlap]
        ys = ys[non_overlap]
        ra = ra[non_overlap]
        dec = dec[non_overlap]
        msg += " ... removed them."
    else:
        msg += " ... keeping them."
    logger.info(msg)

    # Remove all source positions too close to edges of image (where the annulus would
    # extend beyond the image boundaries).
    padding = photometry_apertures.outer_annulus
    out_of_bounds = (
        (xs < padding)
        | (xs > (ccd_image.shape[1] - padding))
        | (ys < padding)
        | (ys > (ccd_image.shape[0] - padding))
    )
    in_bounds = ~out_of_bounds
    # Track dropped sources due to out of bounds positions
    dropped_sources.extend(star_ids[out_of_bounds].tolist())
    # Remove sources too close to the edges
    star_ids = star_ids[in_bounds]
    xs = xs[in_bounds]
    ys = ys[in_bounds]
    ra = ra[in_bounds]
    dec = dec[in_bounds]
    in_cnt = np.sum(in_bounds)
    out_cnt = np.sum(out_of_bounds)
    logger.info(
        f"{logline} {out_cnt} sources too close to image edge ... removed " "them."
    )
    logger.info(
        f"{logline} {in_cnt} of {src_cnt} original sources to have photometry " "done."
    )

    # If we are using x/y positions previously obtained from the ra/dec positions and
    # WCS, then recentroid the sources to refine the positions. This is
    # particularly useful is processing multiple images of the same field
    # and just passing the same sourcelist when calling single_image_photometry
    # on each image.
    if use_coordinates == "sky":
        try:
            xcen, ycen = centroid_sources(
                ccd_image.data, xs, ys, box_size=2 * photometry_apertures.radius + 1
            )
        except NoOverlapError:
            logger.warning(
                f"{logline} Determining new centroids failed ... "
                "SKIPPING THIS IMAGE!"
            )
            return None, None
        else:  # Proceed
            # Calculate offset between centroid in this image and the positions
            # based on input RA/Dec.
            center_diff = np.sqrt((xs - xcen) ** 2 + (ys - ycen) ** 2)

            # The center really shouldn't move more than about the fwhm, could
            # rework this in the future to use that instead.
            too_much_shift = (
                center_diff
                > photometry_settings.source_location_settings.shift_tolerance
            )

            # If the shift is too large, use the WCS-derived positions instead
            # (these sources are probably too faint for centroiding to work well)
            xcen[too_much_shift] = xs[too_much_shift]
            ycen[too_much_shift] = ys[too_much_shift]
            xs, ys = xcen, ycen

    # Compute RA/Dec if not already provided
    if not sourcelist.has_ra_dec:
        try:
            skypos = ccd_image.wcs.pixel_to_world(xs, ys)
            ra = skypos.ra.value
            dec = skypos.dec.value
        except AttributeError:
            ra = [np.nan] * len(xs)
            dec = [np.nan] * len(ys)

    # Define apertures and annuli for the aperture photometry
    aper_locs = np.array([xs, ys]).T
    apers = CircularAperture(aper_locs, r=photometry_apertures.radius)
    annuli = CircularAnnulus(
        aper_locs,
        r_in=photometry_apertures.inner_annulus,
        r_out=photometry_apertures.outer_annulus,
    )

    # Perform the aperture photometry
    photom = aperture_photometry(
        ccd_image.data,
        (apers, annuli),
        mask=ccd_image.mask,
        method=photometry_options.method,
    )

    # Add source ids to the photometry table
    photom["star_id"] = star_ids
    photom["ra"] = ra * u.deg
    photom["dec"] = dec * u.deg

    # Drop ID column from aperture_photometry()
    del photom["id"]

    # Add various CCD image parameters to the photometry table
    if fname is not None:
        photom["file"] = fname
    else:
        photom["file"] = [""] * len(photom)

    # Set various columns based on CCDData headers (which we
    # checked for earlier)
    photom["exposure"] = [exposure] * len(photom) * u.second
    photom["date-obs"] = Time(Column(data=[date_obs]))
    photom["filter"] = [filter] * len(photom)
    photom.rename_column("filter", "passband")

    # Check for airmass keyword in header and set 'airmass' if found,
    # but accept it may not be available
    try:
        photom["airmass"] = [ccd_image.header["AIRMASS"]] * len(photom)
    except KeyError:
        logger.warning(
            f"{logline} 'AIRMASS' not found in CCD " "image header ... setting to NaN!"
        )
        photom["airmass"] = [np.nan] * len(photom)

    # Save aperture and annulus information
    photom.rename_column("aperture_sum_0", "aperture_sum")
    photom.rename_column("aperture_sum_1", "annulus_sum")
    photom["aperture_sum"].unit = ccd_image.unit
    photom["annulus_sum"].unit = ccd_image.unit
    photom["aperture"] = apers.r * u.pixel
    photom["annulus_inner"] = annuli.r_in * u.pixel
    photom["annulus_outer"] = annuli.r_out * u.pixel
    # By convention, area is in units of pixels (not pixels squared) in a digital image
    photom["aperture_area"] = apers.area * u.pixel
    photom["annulus_area"] = annuli.area * u.pixel

    if photometry_options.reject_background_outliers:
        msg = f"{logline} Computing clipped sky stats ... "
        try:
            (
                avg_sky_per_pix,
                med_sky_per_pix,
                std_sky_per_pix,
            ) = clipped_sky_per_pix_stats(ccd_image, annuli)
        except AttributeError:
            msg += "BAD ANNULUS ('sky_per_pix' stats set to np.nan) ... "
            avg_sky_per_pix, med_sky_per_pix, std_sky_per_pix = np.nan, np.nan, np.nan
        photom["sky_per_pix_avg"] = avg_sky_per_pix / u.pixel
        photom["sky_per_pix_med"] = med_sky_per_pix / u.pixel
        photom["sky_per_pix_std"] = std_sky_per_pix / u.pixel
        msg += "DONE."
        logger.info(msg)

    else:  # Don't reject outliers (but why would you do this?)
        logger.warning(
            f"{logline} SUGGESTION: You are computing sky per pixel "
            "without clipping (set reject_background_outliers=True "
            "to perform clipping)."
        )
        med_pp = []
        std_pp = []
        for mask in annuli.to_mask():
            annulus_data = mask.cutout(ccd_image)
            med_pp.append(np.median(annulus_data))
            std_pp.append(np.std(annulus_data))
        photom["sky_per_pix_avg"] = photom["annulus_sum"] / photom["annulus_area"]
        photom["sky_per_pix_med"] = np.array(med_pp) * ccd_image.unit / u.pixel
        photom["sky_per_pix_std"] = np.array(std_pp) * ccd_image.unit / u.pixel

    # Compute counts using clipped stats on sky per pixel
    photom["aperture_net_cnts"] = photom["aperture_sum"].value - (
        photom["aperture_area"].value * photom["sky_per_pix_avg"].value
    )
    photom["aperture_net_cnts"].unit = ccd_image.unit

    # Fit the FWHM of the sources (can result in many warnings due to
    # failed FWHM fitting, capture those warnings and print a summary)
    msg = f"{logline} Fitting FWHM of all sources (may take a few minutes) ... "
    with warnings.catch_warnings(record=True) as warned:
        warnings.filterwarnings("always", category=AstropyUserWarning)
        fwhm_x, fwhm_y = compute_fwhm(
            ccd_image,
            photom,
            fwhm_estimate=photometry_apertures.fwhm,
            fit=photometry_options.fwhm_by_fit,
        )
        num_warnings = len(warned)
        msg += f"fitting failed on {num_warnings} of {len(photom)} sources  ... "
    msg += "DONE."
    logger.info(msg)

    # Deal with bad FWHM values
    bad_fwhm = (fwhm_x < 1) | (fwhm_y < 1)  # Set bad values to NaN now
    fwhm_x[bad_fwhm] = np.nan
    fwhm_y[bad_fwhm] = np.nan
    photom["fwhm_x"] = fwhm_x * u.pixel
    photom["fwhm_y"] = fwhm_y * u.pixel
    photom["width"] = ((fwhm_x + fwhm_y) / 2) * u.pixel
    if np.sum(bad_fwhm) > 0:
        logger.info(
            f"{logline} Bad FWHM values (<1 pixel) for {np.sum(bad_fwhm)} " "sources."
        )

    # Flag sources with bad counts before computing noise.
    # This can happen, for example, when the object is faint and centroiding is
    # bad.  It can also happen when the sky background is low.
    bad_cnts = photom["aperture_net_cnts"].value < 0
    # This next line works because booleans are just 0/1 in numpy
    if np.sum(bad_cnts) > 0:
        logger.info(
            f"{logline} Aperture net counts negative for {np.sum(bad_cnts)} " "sources."
        )

    all_bads = bad_cnts | bad_fwhm

    photom["aperture_net_cnts"][all_bads] = np.nan
    logger.info(
        f"{logline} {np.sum(all_bads)} sources with either bad FWHM fit "
        "or bad aperture net counts had aperture_net_cnts set to NaN."
    )

    # Compute instrumental magnitudes
    photom["mag_inst"] = -2.5 * np.log10(
        camera.gain.value * photom["aperture_net_cnts"].value / photom["exposure"].value
    )

    # Compute and save noise
    msg = f"{logline} Calculating noise for all sources ... "
    noise = calculate_noise(
        camera=camera,
        counts=photom["aperture_net_cnts"].value,
        sky_per_pix=photom["sky_per_pix_avg"].value,
        aperture_area=photom["aperture_area"].value,
        annulus_area=photom["annulus_area"].value,
        exposure=photom["exposure"].value,
        include_digitization=photometry_options.include_dig_noise,
    )
    photom["noise_electrons"] = noise  # Noise in electrons
    photom["noise_electrons"].unit = u.electron
    photom["noise_cnts"] = noise / camera.gain.value  # Noise in counts
    photom["noise_cnts"].unit = ccd_image.unit

    # Compute and save SNR
    snr = camera.gain.value * photom["aperture_net_cnts"] / noise
    photom["snr"] = snr
    photom["mag_error"] = 1.085736205 / snr
    msg += "DONE."
    logger.info(msg)

    # Close logfile if it was created
    if fh_exists:
        fh.flush()
        fh.close()

    # Remove logger handler
    logger.handlers.clear()

    # Create PhotometryData object to return
    photom_data = PhotometryData(
        observatory=observatory,
        camera=camera,
        input_data=photom,
        passband_map=passband_map,
    )

    return photom_data, dropped_sources


def multi_image_photometry(
    directory_with_images,
    photometry_settings,
    reject_unmatched=True,
    object_of_interest=None,
):
    """
    Perform aperture photometry on a directory of images.

    Parameters
    ----------

    directory_with_images : str
        Folder containing the images on which to do photometry. Photometry
        will only be done on images that contain the ``object_of_interest``.
        All images *must* have WCS headers and the following headers: OBJECT,
        DATE-OBS, an exposure time header (which can be any of the following: EXPOSURE,
        EXPTIME, TELAPSE, ELAPTIME, ONTIME, or LIVETIME), and FILTER.  If AIRMASS is
        available it will be added to `phot_table`.

    photometry_settings : `stellarphot.settings.PhotometrySettings`
        Photometry settings to use for the photometry.  This includes the camera,
        observatory, the aperture and annulus radii to use for the photometry, a map
        of passbands from your filters to AAVSO filter names, and options for the
        photometry. See `stellarphot.settings.PhotometrySettings` for more information.

    reject_unmatched : bool, optional (Default: True)
        If ``True``, any sources that are not detected on all the images are
        rejected.  If you are interested in a source that can intermittently
        fall below your detection limits, we suggest setting this to ``False``
        so that all sources detected on each image are reported.

    object_of_interest : str, optional (Default: None)
        Name of the object of interest. The only files on which photometry
        will be done are those whose header contains the keyword ``OBJECT``
        whose value is ``object_of_interest``. *Only used for multi-image
        photometry* to select which files to perform photometry on.

    Returns
    -------

    phot_table : `stellarphot.PhotometryData`
        Photometry data for all the sources on which aperture photometry was
        performed in all the images. This may be a subset of the sources in
        the sourcelist if locations were too close to the edge of any one image
        or to each other for successful aperture photometry.

    """
    sourcelist = SourceListData.read(
        photometry_settings.source_location_settings.source_list_file
    )

    # Initialize lists to track all PhotometryData objects and all dropped sources
    phots = []
    missing_sources = []

    # Confirm sourcelist has ra/dec coordinates
    if not sourcelist.has_ra_dec:
        raise ValueError(
            "multi_image_photometry: sourcelist must have RA/Dec "
            "coordinates to use this function."
        )

    # Set up logging (retrieve a logger but purge any existing handlers)
    multilogger = logging.getLogger("multi_image_photometry")
    multilogger.setLevel(logging.INFO)
    console_format = logging.Formatter("%(message)s")
    # Remove all other existing handlers from the logger
    # (Kind of brute force, but works for our purposes
    for handler in multilogger.handlers[:]:
        multilogger.removeHandler(handler)

    logfile = photometry_settings.logging_settings.logfile
    console_log = photometry_settings.logging_settings.console_log

    # Set up logging:
    # Check if logfile is not None, set up logging to be written to the logfile.
    # Next check if console_log is True, if it is, set up logging to be written
    #    to the console.
    # If neither of these are true, set up logging to be written to NullHandler
    #    which effectively suppresses the logging output.

    # Set up logging to a file (in addition to any logging below)
    if logfile is not None:
        # Get the name of the logfile without the path
        logfile = Path(logfile).name
        # Redirect the logfile to the directory_with_images
        logfile = Path(directory_with_images) / logfile
        # Change the settings so when they are passed to single_image_photometry
        # the logging will be written to the same logfile
        photometry_settings.logging_settings.logfile = str(logfile)

        # Keep original name without path
        logfile_name = str(Path(logfile).name)

        # by default this appends to existing logfile
        fh = logging.FileHandler(logfile)
        log_format = logging.Formatter("%(levelname)s - %(message)s")
        fh.setFormatter(log_format)
        fh.setLevel(logging.INFO)
        multilogger.addHandler(fh)

    # Set up logging to console if requested
    if console_log:
        ch = logging.StreamHandler()
    else:  # otherwise effectively suppress output
        ch = logging.NullHandler()
    ch.setFormatter(console_format)
    ch.setLevel(logging.INFO)
    multilogger.addHandler(ch)

    ##
    ## Process all the individual files
    ##

    # Build image file collection
    ifc = ImageFileCollection(directory_with_images)

    # Disable any root logger handlers that are active before using
    # logging (must be done here because ImageFileCollection() creates
    # a logger)
    if logging.root.hasHandlers():
        logging.root.handlers.clear()

    n_files_processed = 0

    msg = f"Starting photometry of files in {directory_with_images} ... "
    if logfile is not None:
        msg += f"logging output to {logfile_name}"
        # If not logging to console, print message here
        if not console_log:
            print(msg)
    multilogger.info(msg)

    # Suppress the FITSFixedWarning that is raised when reading a FITS file header
    warnings.filterwarnings("ignore", category=FITSFixedWarning)

    # Process all the files
    for this_ccd, this_fname in ifc.ccds(object=object_of_interest, return_fname=True):
        multilogger.info(f"multi_image_photometry: Processing image {this_fname}")
        if this_ccd.wcs is None:
            multilogger.warning("                   .... SKIPPING THIS IMAGE (NO WCS)")
            continue

        # Call single_image_photometry on each image
        n_files_processed += 1
        multilogger.info("  Calling single_image_photometry ...")
        this_phot, this_missing_sources = single_image_photometry(
            this_ccd,
            photometry_settings,
            fname=this_fname,
            logline="    >",
        )
        if (this_phot is None) or (this_missing_sources is None):
            multilogger.info("  single_image_photometry failed for this image.")
        else:
            multilogger.info(
                f"  Done with single_image_photometry for {this_fname}\n\n"
            )

            # Extend the list of missing stars
            missing_sources.extend(this_missing_sources)

            # And add the final table to the list of tables
            phots.append(this_phot)

    if n_files_processed == 0:
        raise RuntimeError("No images were processed!")

    ##
    ## Done processing individual images, now combine them into one table
    ##

    # Combine all of the individual photometry tables into one.
    # Attributes should survive intact assume all have same camera and observatory.
    all_phot = vstack(phots)

    # If requested, eliminate source not detected on every image by building
    # a set of all the unique star_ids that were missing on at least one image.
    if reject_unmatched and len(missing_sources) > 0:
        if len(missing_sources) > 1:
            uniques = set(missing_sources)
        else:
            uniques = set([missing_sources])

        msg = f"  Removing {len(uniques)} sources not observed in every image ... "
        # Purge the photometry table of all sources that were eliminated
        # on at least one image
        starid_to_remove = sorted([u for u in uniques if u in all_phot["star_id"]])
        # add index to PhotometryData to speed up removal
        all_phot.add_index("star_id")
        # Remove the starid for objects not observed in every image
        if starid_to_remove:
            bad_rows = all_phot.loc_indices[starid_to_remove]
            try:
                bad_rows = list(bad_rows)
            except TypeError:
                bad_rows = [bad_rows]
            all_phot.remove_indices("star_id")
            all_phot.remove_rows(sorted(bad_rows))
        # Drop index from PhotometryData to save memory
        all_phot.remove_indices("star_id")
        msg += "DONE."
        multilogger.info(msg)

    multilogger.info(
        f"  DONE processing all matching images in {directory_with_images}"
    )
    if logfile is not None and not console_log:
        print(f"  DONE processing all matching images in {directory_with_images}")

    # Close the logfile if it is open
    if logfile is not None:
        fh.flush()
        fh.close()
    # Remove logger handler
    multilogger.handlers.clear()

    return all_phot


def faster_sigma_clip_stats(data, sigma=5, iters=5, axis=None):
    """
    Calculate sigma clipped stats quickly using NaNs instead of masking
    and using bottleneck where possible.

    Parameters
    ----------

    data : numpy array
        The data to be clipped. *The data should have had masked values
        replaced with NaN prior to calling this function.*

    sigma : float, optional
        Number of standard deviations (estimated with the MAD) a point must
        be from the central value (median) to be rejected.

    iters : int, optional
        Number of sigma clipping iterations to perform. Fewer iterations than
        this may be performed because iterations stop when no new data is
        being clipped.

    axis : int, optional
        axis along which to perform the median.

    Returns
    -------
    mean, median, std : float or numpy array
        Clipped statistics; shape depends on the shape of the input.
    """
    data = data.copy()
    for _ in range(iters):
        central = bn.nanmedian(data, axis=axis)
        try:
            central = central[:, np.newaxis]
        except (ValueError, IndexError, TypeError):
            pass

        std_dif = 1.4826 * bn.nanmedian(np.abs(data - central))

        clips = np.abs(data - central) / std_dif > sigma

        if np.nansum(clips) == 0:
            break
        data[clips] = np.nan
    return (
        bn.nanmean(data, axis=axis),
        bn.nanmedian(data, axis=axis),
        bn.nanstd(data, axis=axis),
    )


def find_too_close(sourcelist, aperture_rad, pixel_scale=None):
    """
    Identify sources that are closer together than twice the aperture radius.

    If 'xcenter' and 'ycenter' are available in the sourcelist (as determined by
    the value of sourcelist.has_x_y), they are used to determine separation of sources
    in units of pixels, otherwise if only ra/dec are available, the pixel_scale
    is necessary to determine the separation.

    Parameters
    ----------

    sourcelist : `stellarphot.SourceListData`
        A list of sources with x/y and/or RA/Dec coordinates.

    aperture_rad : int
        Radius of the aperture, in pixels.

    pixel_scale : `float, optional` (Default: None)
        Pixel scale of the image in arcsec/pixel. Only required
        if x/y coordinates are NOT provided.

    Returns
    -------

    numpy array of bool
        Array the same length as the RA/Dec that is ``True`` where the sources
        are closer than two aperture radii, ``False`` otherwise.
    """
    if not isinstance(sourcelist, SourceListData):
        raise TypeError(
            "sourcelist must be of type SourceListData not " f"'{type(sourcelist)}'"
        )

    if not isinstance(pixel_scale, float):
        raise TypeError(f"pixel_scale must be a float not '{type(pixel_scale)}'")

    if sourcelist.has_x_y:
        x, y = sourcelist["xcenter"], sourcelist["ycenter"]
        # Find the pixel distance to the nearest neighbor for each source
        dist_mat = cdist(np.array([x, y]).T, np.array([x, y]).T, metric="euclidean")
        np.fill_diagonal(dist_mat, np.inf)
        # Return array with True where the distance is less than twice the aperture
        # radius
        return dist_mat.min(0) < 2 * aperture_rad
    elif sourcelist.has_ra_dec:
        if pixel_scale is None:
            raise ValueError(
                "pixel_scale must be provided if x/y coordinates are "
                "not available in the sourcelist."
            )
        star_coords = SkyCoord(
            ra=sourcelist["ra"], dec=sourcelist["dec"], frame="icrs", unit="degree"
        )
        idxc, d2d, d3d = star_coords.match_to_catalog_sky(star_coords, nthneighbor=2)
        return d2d < (aperture_rad * 2 * pixel_scale * u.arcsec)
    else:
        raise ValueError("sourcelist must have x/y or ra/dec coordinates")


def clipped_sky_per_pix_stats(data, annulus, sigma=5, iters=5):
    """
    Calculate sigma-clipped statistics on an annulus.

    Parameters
    ----------

    data : `astropy.nddata.CCDData`
        CCD image on which the annuli are defined.

    annulus : `photutils.CircularAnnulus`
        One or more annulus (of any shape) from photutils.

    sigma : float, optional
        Number of standard deviations from the central value a
        pixel must be to reject it.

    iters : int, optional
        Maximum number of sigma clip iterations to perform. Iterations stop
        automatically if no pixels are rejected.

    Returns
    -------

    avg_sky_per_pix, med_sky_per_pix, std_sky_per_pix : `astropy.units.Quantity`
        Average, median and standard deviation of the sky per pixel.

    """
    # Get a list of masks from the annuli
    # Use the 'center' method because then pixels are either in or out. To use
    # 'partial' or 'exact' we would need to do a weighted sigma clip and
    # I'm not sure how to do that.
    masks = annulus.to_mask(method="center")

    anul_list = []
    for mask in masks:
        # Multiply the mask times the data
        to_clip = mask.multiply(data.data, fill_value=np.nan)
        anul_list.append(to_clip.flatten())
    # Convert the list to an array for doing the sigma clip
    anul_array = np.array(anul_list)
    # Turn all zeros into np.nan...
    anul_array[anul_array == 0] = np.nan
    avg_sky_per_pix, med_sky_per_pix, std_sky_per_pix = faster_sigma_clip_stats(
        anul_array, sigma=sigma, iters=iters, axis=1
    )

    return (
        avg_sky_per_pix * data.unit,
        med_sky_per_pix * data.unit,
        std_sky_per_pix * data.unit,
    )


def calculate_noise(
    camera=None,
    counts=0.0,
    sky_per_pix=0.0,
    aperture_area=0,
    annulus_area=0,
    exposure=0,
    include_digitization=False,
):
    """
    Computes the noise in a photometric measurement.

    This function computes the noise (in units of electrons) in a photometric
    measurement using the revised CCD equation from Collins et al (2017) AJ, 153, 77
    who based their expression on the one originally proposed for SNR by
    Merline, W. J., & Howell, S. B. 1995, Experimental Astronomy, 6, 163.

    The equation is:

    .. math::

        \\sigma = \\sqrt{G \\cdot N_C + A \\cdot \\left(1 + \\frac{A}{B}\\right)\\cdot
        \\left[ G\\cdot S + D \\cdot t + R^2 + (0.289 G)^2\\right]}

    where :math:`\\sigma` is the noise, :math:`G` is the gain,
    :math:`N_C` is the source counts (which is photon/electron counts divided by gain),
    :math:`A` is the aperture area in pixels,
    :math:`B` is the annulus area in pixels,
    :math:`S` is the sky counts per pixel,
    :math:`D` is the dark current in electrons per second,
    :math:`R` is the read noise in electrons per pixel per read,
    and :math:`t` is exposure time in seconds.

    Note: The :math:`(0.289 G)^2` term is "digitization noise" and is optional.

    Parameters
    ----------

    camera : `stellarphot.Camera`
        The camera object that contains the gain, read noise and dark current
        of the CCD.

    counts : float, optional
        Counts of the source.

    sky_per_pix : float, optional
        Sky per pixel in counts per pixel.

    aperture_area : int, optional
        Area of the aperture in pixels.

    annulus_area : int, optional
        Area of the annulus in pixels.

    exposure : int, optional
        Exposure time in seconds.

    include_digitization : bool, optional
        Whether to include the digitization noise. Defaults to False.

    Returns
    -------

    noise : float
        The noise in the photometric measurement in electrons.
    """
    if camera is None:
        raise ValueError("camera must be provided")
    elif not isinstance(camera, Camera):
        raise ValueError(f"camera must be of type Camera not '{type(camera)}'")

    # Extract values from camera object
    gain = camera.gain.value
    dark_current_per_sec = camera.dark_current.value
    read_noise = camera.read_noise.value

    try:
        no_annulus = (annulus_area == 0).all()
    except AttributeError:
        no_annulus = annulus_area == 0

    if no_annulus:
        area_ratio = aperture_area
    else:
        area_ratio = aperture_area * (1 + aperture_area / annulus_area)

    # Convert counts to electrons
    poisson_source = gain * counts

    try:
        poisson_source = poisson_source.value
    except AttributeError:
        pass

    sky = area_ratio * gain * sky_per_pix
    dark = area_ratio * dark_current_per_sec * exposure
    rn_error = area_ratio * read_noise**2

    digitization = 0.0

    if include_digitization:
        digitization = area_ratio * (gain * 0.289) ** 2

    return np.sqrt(poisson_source + sky + dark + rn_error + digitization)



from stellarphot.settings import Camera, SourceLocationSettings, ui_generator
from stellarphot.settings.tests.test_models import TEST_CAMERA_VALUES


class TestUiGenerator:
    def test_camera(self):
        ui = ui_generator(Camera)
        # The description should be the beginning of the docstring
        assert Camera.__doc__.strip().startswith(ui.description.split()[0])

        # We always want to show nullable fields
        assert ui.show_null

        # Which means we don't need the button to show/hide them
        assert ui.bn_shownull.layout.display == "none"

        # For now we do not show the validation output because it is painfully
        # verbose and not very helpful.
        assert not ui.show_validation

        # We always display nested models
        assert ui.open_nested

        # Finally, we don't want the widget to update continuously because it
        # will overwrite the value the user entered.
        for widget in ui.di_widgets.values():
            if hasattr(widget, "continuous_update"):
                assert not widget.continuous_update

    def test_disabled_state_save_button(self):
        # We want the save button to be enabled only when
        # 1. The user has made a change, AND
        # 2. The value in the widget is a valid pydantic model

        ui = ui_generator(Camera)
        # The save button should be disabled
        assert ui.savebuttonbar.bn_save.disabled

        # Set one field to a valid value....
        ui.value["name"] = "test"

        # We need to manually indicate a change has been made
        ui.savebuttonbar.unsaved_changes = True

        # ...and the save button should still be disabled, because
        # the rest of the fields are not valid.
        assert ui.savebuttonbar.bn_save.disabled

        # Set a valid value
        ui.value = TEST_CAMERA_VALUES

        # So it turns out that the validation stuff only updates when changes are made
        # in the UI rather than programmatically. Since we know we've set a valid value,
        # and that we've made changes we just manually set the relevant values.
        ui.savebuttonbar.unsaved_changes = True
        ui.is_valid.value = True

        # The save button should now be enabled
        assert not ui.savebuttonbar.bn_save.disabled

        # Click on save
        ui.savebuttonbar.bn_save.click()

        # Unsaved changes should be False
        assert not ui.savebuttonbar.unsaved_changes

        # The save and revert buttons should now be disabled
        assert ui.savebuttonbar.bn_save.disabled

    def test_disabled_state_revert_button(self):
        # We want the revert button to be enabled whenever the user has made a change,
        # regardless of whether the value is valid or not.

        ui = ui_generator(Camera)
        # The revert button should be disabled
        assert ui.savebuttonbar.bn_revert.disabled

        # Set one field to a valid value -- the rest of the fields are still invalid
        ui.value["name"] = "test"

        # We need to manually indicate a change has been made
        ui.savebuttonbar.unsaved_changes = True

        # ...and the revert button should be enabled
        assert not ui.savebuttonbar.bn_revert.disabled

        # Click on revert
        ui.savebuttonbar.bn_revert.click()

        # The revert button should now be disabled
        assert ui.savebuttonbar.bn_revert.disabled

        # Unsaved changes should be False
        assert not ui.savebuttonbar.unsaved_changes

    def test_field_width(self):
        # We want to set the width of all except FileChooser fields to a fixed value

        # Nothing special about the value below except that it is unlikely to be
        # a default...
        width = "142px"
        # Use SourceLocationSettings because it has a FileChooser field.
        ui = ui_generator(SourceLocationSettings, max_field_width=width)

        for widget in ui.di_widgets.values():
            if widget.__class__.__name__ == "FileChooser":
                # FileChooser widgets have a different layout object
                assert widget.layout.max_width != width
            else:
                assert widget.layout.max_width == width

    def test_file_chooser_width(self):
        # We want to set the width of FileChooser fields, but no other fields,
        # to a fixed value.

        # Nothing special about the value below except that it is unlikely to be
        # a default...
        width = "142px"
        # Use SourceLocationSettings because it has a FileChooser field.
        ui = ui_generator(SourceLocationSettings, file_chooser_max_width=width)

        for widget in ui.di_widgets.values():
            if widget.__class__.__name__ == "FileChooser":
                assert widget.layout.max_width == width
            else:
                assert widget.layout.max_width != width



import os
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest

from stellarphot.settings import (
    SETTINGS_FILE_VERSION,
    Camera,
    Observatory,
    PartialPhotometrySettings,
    PassbandMap,
    PhotometrySettings,
    PhotometryWorkingDirSettings,
    SavedSettings,
    settings_files,  # This import is needed for mocking -- see TestSavedSettings
)
from stellarphot.settings.tests.test_models import TEST_PHOTOMETRY_SETTINGS

CAMERA = """
{
    "name": "Aspen CG 16m",
    "data_unit": "adu",
    "gain": "1.5 electron / adu",
    "read_noise": "10.0 electron",
    "dark_current": "0.01 electron / s",
    "pixel_scale": "0.6 arcsec / pix",
    "max_data_value": "50000.0 adu"
}
"""

OBSERVATORY = """
{
    "name": "Feder",
    "latitude": "46d52m25.68s",
    "longitude": "263d13m55.92s",
    "elevation": "311.0 m",
    "AAVSO_code": null,
    "TESS_telescope_code": null
}
"""

PASSBAND_MAP = """
{
    "name": "Filter wheel 1",
    "your_filter_names_to_aavso": [
        {
            "your_filter_name": "rp",
            "aavso_filter_name": "SR"
        },
        {
            "your_filter_name": "gp",
            "aavso_filter_name": "SG"
        }
    ]
}
"""


# Keep this test out of the class so that it uses the real settings path.
def test_settings_path_contains_package_and_version():
    # Make sure that the path to the settings file contains the package name and
    # version.
    saved_settings = SavedSettings(_create_path=False)
    assert "stellarphot" in str(saved_settings.settings_path)
    assert SETTINGS_FILE_VERSION in str(saved_settings.settings_path)


class TestSavedSettings:
    # This pytest fixture is used to create a fake settings directory for the tests.

    # Being a fixture means it can be passed into tests or other functions, just like
    # the fixture tmp_path.
    # The autouse=True parameter means that this fixture will be provided to every test
    # in this class without needing to be explicitly passed in.
    @pytest.fixture(autouse=True)
    def fake_settings_dir(self, mocker, tmp_path):
        # mocker is a pytest fixture provided by the pytest-mock package. It is used to
        # mock objects and functions.
        # Mocking means providing a fake version of an object, function, attribute, or
        # method that can be used in place of the real thing.

        # One of the confusing things is figuring out what to mock. In this case, we are
        # mocking the user_data_dir attribute of the PlatformDirs class in the
        # settings_files module. To make sure that is the PlatformDirs class we are
        # mocking, we need to specifically mock settings_files.PlatformDirs. A few
        # things that wouldn't work, for example, are importing PlatformsDirs directly
        # from platformdirs in this module and then trying to mock that, or importing
        # PlatformDirs from settings_files and then trying to mock that. Actually,
        # that last thing might work, but there is some values in being explicit here.
        # doing it that way does mean importing the settings_files module.
        #
        # This attribute is used to determine the path to the
        # settings directory. By mocking it, we can control where the settings directory
        # is created and use a temporary directory for the tests.

        # stellarphot is added to the name of the directory to make sure we start
        # without a stellarphot directory for each test.
        mocker.patch.object(
            settings_files.PlatformDirs, "user_data_dir", tmp_path / "stellarphot"
        )

    def test_settings_path_is_created_if_not_exists(self):
        # Check that the settings path is created if it doesn't exist.
        # It is important to use settings_files.PlatformDirs instead
        # of, say, importing PlatformDirs directly because we want to use the mocked
        # version of the attribute.
        assert not Path(settings_files.PlatformDirs.user_data_dir).exists()
        saved_settings = SavedSettings()
        assert saved_settings.settings_path.exists()

    @pytest.mark.parametrize(
        "klass,item_json",
        [(Camera, CAMERA), (Observatory, OBSERVATORY), (PassbandMap, PASSBAND_MAP)],
    )
    def test_add_saved_item(self, klass, item_json):
        # Test that items are properly saved and loaded.
        saved_settings = SavedSettings()
        assert saved_settings.settings_path.exists()
        # Add a camera.
        item = klass.model_validate_json(item_json)
        saved_settings.add_item(item)
        # Load the cameras
        saved_items = saved_settings.get_items(item)
        assert len(saved_items.as_dict) == 1
        assert saved_items.as_dict[item.name] == item

    @pytest.mark.parametrize(
        "klass,item_json",
        [(Camera, CAMERA), (Observatory, OBSERVATORY), (PassbandMap, PASSBAND_MAP)],
    )
    def test_adding_multiple_items_of_same_type(self, klass, item_json):
        # Test that items are properly saved and loaded.
        saved_settings = SavedSettings()
        assert saved_settings.settings_path.exists()
        # Add an item
        item1 = klass.model_validate_json(item_json)
        item2 = klass.model_validate_json(
            item_json.replace(item1.name, item1.name + "2")
        )
        saved_settings.add_item(item1)
        saved_settings.add_item(item2)
        # Load the items -- any instance of the class (e.g. Camera) or a
        # string (e.g. "camera") should work for getting the items.
        saved_items = saved_settings.get_items(item1)
        assert len(saved_items.as_dict) == 2
        assert saved_items.as_dict[item2.name] == item2
        assert saved_items.as_dict[item1.name] == item1

    def test_add_existing_saved_item_raises_error(self):
        # Test that adding an existing camera raises an error. Other items follow the
        # same pattern, so only cameras are tested.
        saved_settings = SavedSettings()
        # Add a camera.
        item = Camera.model_validate_json(CAMERA)
        saved_settings.add_item(item)
        # Add the same camera again.
        with pytest.raises(
            ValueError, match="Aspen CG 16m already exists in cameras.json"
        ):
            saved_settings.add_item(item)

    def test_adding_multiple_types_of_items(self):
        # Test that adding multiple types of items works.
        saved_settings = SavedSettings()
        # Add a camera.
        camera = Camera.model_validate_json(CAMERA)
        saved_settings.add_item(camera)
        # Add an observatory.
        observatory = Observatory.model_validate_json(OBSERVATORY)
        saved_settings.add_item(observatory)
        # Add a passband map.
        passband_map = PassbandMap.model_validate_json(PASSBAND_MAP)
        saved_settings.add_item(passband_map)
        # Load the items
        cameras = saved_settings.get_items("camera")
        assert len(cameras.as_dict) == 1
        assert cameras.as_dict[camera.name] == camera
        observatories = saved_settings.get_items("observatory")
        assert len(observatories.as_dict) == 1
        assert observatories.as_dict[observatory.name] == observatory
        passband_maps = saved_settings.get_items("passband_map")
        assert len(passband_maps.as_dict) == 1
        assert passband_maps.as_dict[passband_map.name] == passband_map

    def test_delete_without_confirm_raises_error(self):
        # Trying to delete settings without confirming should raise an error.
        saved_settings = SavedSettings()
        with pytest.raises(ValueError, match="You must confirm deletion by passing"):
            saved_settings.cameras.delete()

    def test_delete_with_confirm_deletes_file(self):
        # Test that deleting a settings file works.
        saved_settings = SavedSettings()
        # Add a camera.
        camera = Camera.model_validate_json(CAMERA)
        saved_settings.add_item(camera)
        # Delete the camera.
        saved_settings.cameras.delete(confirm=True)
        assert not (
            saved_settings.settings_path / saved_settings.cameras._file_name
        ).exists()

    def test_deleting_all_settings_without_confirm_raises_error(self):
        # Trying to delete all settings without confirming should raise an error.
        saved_settings = SavedSettings(_create_path=False)
        with pytest.raises(ValueError, match="You must confirm deletion by passing"):
            saved_settings.delete()

    def test_delete_all_settings_with_confirm_deletes_files(self):
        # Test that deleting all settings files works.
        saved_settings = SavedSettings()
        # Add a camera.
        camera = Camera.model_validate_json(CAMERA)
        saved_settings.add_item(camera)
        # Add an observatory.
        observatory = Observatory.model_validate_json(OBSERVATORY)
        saved_settings.add_item(observatory)
        # Add a passband map.
        passband_map = PassbandMap.model_validate_json(PASSBAND_MAP)
        saved_settings.add_item(passband_map)

        # Delete all settings.
        saved_settings.delete(confirm=True, delete_settings_folder=True)
        assert not (
            saved_settings.settings_path / saved_settings.cameras._file_name
        ).exists()
        assert not (
            saved_settings.settings_path / saved_settings.observatories._file_name
        ).exists()
        assert not (
            saved_settings.settings_path / saved_settings.passband_maps._file_name
        ).exists()
        assert not saved_settings.settings_path.exists()

    def test_delete_all_with_no_settings_works(self):
        # Test that deleting all settings files works when no settings are present.
        saved_settings = SavedSettings()
        # Delete all settings but not the settings folder
        saved_settings.delete(confirm=True)
        assert len(list(saved_settings.settings_path.glob("*"))) == 0

        # Delete all settings and the settings folder
        saved_settings.delete(confirm=True, delete_settings_folder=True)
        assert not saved_settings.settings_path.exists()

    def test_delete_item_from_collection_works(self):
        # Test that deleting an item from a collection works.
        saved_settings = SavedSettings()
        # Add a camera.
        camera = Camera.model_validate_json(CAMERA)
        saved_settings.add_item(camera)
        camera2 = Camera.model_validate_json(CAMERA.replace("Aspen CG 16m", "foo"))
        saved_settings.add_item(camera2)

        # Make sure both cameras are in the collection.
        assert len(saved_settings.cameras.as_dict) == 2
        # Delete the second camera.
        saved_settings.cameras.delete(name=camera2.name, confirm=True)
        assert len(saved_settings.cameras.as_dict) == 1

    def test_delete_item_from_collection_with_unknown_item_fails(self):
        # Test that trying to delete an unknown item from a collection fails.
        saved_settings = SavedSettings()
        camera = Camera.model_validate_json(CAMERA)
        saved_settings.add_item(camera)
        with pytest.raises(ValueError, match="not found in"):
            saved_settings.cameras.delete(name=camera.name + "foo", confirm=True)

    def test_revtrieving_item_by_name_works(self):
        # Test that retrieving an item by name works.
        saved_settings = SavedSettings()
        # Add a camera.
        camera = Camera.model_validate_json(CAMERA)
        saved_settings.add_item(camera)
        # Retrieve the camera by name.
        retrieved_camera = saved_settings.cameras.get(camera.name)
        assert retrieved_camera == camera

    def test_get_item_with_unknown_item_fails(self):
        # Test that trying to get an unknown item fails.
        saved_settings = SavedSettings()
        with pytest.raises(ValueError, match="Unknown item foo of type"):
            saved_settings.get_items("foo")

    def test_add_item_with_unknown_item_fails(self):
        # Test that trying to add an unknown item fails.
        saved_settings = SavedSettings()
        with pytest.raises(ValueError, match="Unknown item foo of type"):
            saved_settings.add_item("foo")

    @pytest.mark.parametrize(
        "klass,item_json",
        [(Camera, CAMERA), (Observatory, OBSERVATORY), (PassbandMap, PASSBAND_MAP)],
    )
    def test_saved_settings_delete_item(self, klass, item_json):
        # Test that items can be deleted.
        saved_settings = SavedSettings()
        # Add item.
        item = klass.model_validate_json(item_json)
        saved_settings.add_item(item)
        # Verify that the item is there
        assert saved_settings.get_items(klass.__name__).as_dict[item.name] == item

        saved_settings.delete_item(item, confirm=True)
        # Verify that the item was deleted.
        assert len(saved_settings.get_items(klass.__name__).as_dict) == 0

    def test_saved_settings_delete_item_with_unknown_item_fails(self):
        # Test that trying to delete an unknown item fails.
        saved_settings = SavedSettings()
        with pytest.raises(ValueError, match="Unknown item foo of type"):
            saved_settings.delete_item("foo", confirm=True)

    def test_saved_settings_delete_item_with_confirm_false_fails(self):
        # Test that trying to delete an item without confirming fails.
        saved_settings = SavedSettings()
        # Make a camera and save it
        camera = Camera.model_validate_json(CAMERA)
        saved_settings.add_item(camera)
        with pytest.raises(ValueError, match="You must confirm deletion by passing"):
            saved_settings.delete_item(camera, confirm=False)

    def test_saved_settings_delete_item_valid_item_not_in_collection_fails(self):
        # Test that trying to delete an item that is not in the collection fails.
        saved_settings = SavedSettings()
        # Make a camera but don't save it
        camera = Camera.model_validate_json(CAMERA)
        with pytest.raises(ValueError, match="not found in"):
            saved_settings.delete_item(camera, confirm=True)

    def test_saved_settings_round_trip_with_unicode_name(self):
        # Test that items with unicode names can be saved and loaded.
        saved_settings = SavedSettings()
        # Add a camera. This particular name causes a failure on Windows because the
        # default encoding doesn't include Korean characters.
        camera_name = "크레이그"
        camera = Camera(
            name=camera_name,
            data_unit="adu",
            gain="1.5 electron / adu",
            read_noise="10.0 electron",
            dark_current="0.01 electron / s",
            pixel_scale="0.6 arcsec / pix",
            max_data_value="50000.0 adu",
        )
        saved_settings.add_item(camera)
        # Load the camera.
        loaded_camera = saved_settings.get_items("camera").as_dict[camera_name]
        assert loaded_camera == camera


class TestPhotometryWorkingDirSettings:
    def setup_class(cls):
        cls.temp_dir = TemporaryDirectory()

    def teardown_class(cls):
        cls.temp_dir.cleanup()

    def setup_method(self, _):
        # Need to accept the second argument, but don't use it.
        self.original_wdir = Path.cwd()
        os.chdir(self.temp_dir.name)
        for file in Path.cwd().glob("*.json"):
            file.unlink()

    def teardown_method(self, _):
        os.chdir(self.original_wdir)

    def test_sneaky_name_not_accepted(self):
        sneaky_names = [
            "../myfile.json",
            "/some/absolute/path",
            "file_with_no_json_extension.txt",
            "file_with_no_extension",
            " started with a space",
        ]
        for name in sneaky_names:
            with pytest.raises(ValueError, match="not a valid name. The name can"):
                PhotometryWorkingDirSettings(settings_file_name=name)

    def test_bad_settings_value_raises_error(self):
        settings_file = PhotometryWorkingDirSettings()
        error_message = (
            "Settings must be PhotometrySettings or PartialPhotometrySettings"
        )
        with pytest.raises(ValueError, match=error_message):
            settings_file.save("foo")

    def test_save_partial_settings(self):
        # Test that saving partial settings works.
        settings_file = PhotometryWorkingDirSettings()
        settings = PartialPhotometrySettings()
        settings_file.save(settings)
        assert settings_file.partial_settings_file.exists()
        assert not settings_file.settings_file.exists()
        assert settings_file.partial_settings == settings

    def test_save_complete_settings(self):
        # Test that saving complete settings works.
        settings_file = PhotometryWorkingDirSettings()
        settings = PhotometrySettings(**TEST_PHOTOMETRY_SETTINGS)
        settings_file.save(settings)
        assert settings_file.settings_file.exists()
        assert not settings_file.partial_settings_file.exists()
        assert settings_file.settings == settings

    def test_save_partial_settings_that_are_full_settings(self):
        # Test that saving partial settings that are actually full settings works.
        settings_file = PhotometryWorkingDirSettings()
        settings = PartialPhotometrySettings(**TEST_PHOTOMETRY_SETTINGS)
        settings_file.save(settings)
        assert settings_file.settings_file.exists()
        assert not settings_file.partial_settings_file.exists()
        assert settings_file.settings == settings
        assert settings_file.partial_settings is None

    def test_save_partial_then_full_settings(self):
        # Test that saving partial settings and then full settings works.
        settings_file = PhotometryWorkingDirSettings()
        partial_settings = PartialPhotometrySettings()
        settings_file.save(partial_settings)
        assert settings_file.partial_settings_file.exists()
        assert not settings_file.settings_file.exists()
        assert settings_file.partial_settings == partial_settings

        full_settings = PhotometrySettings(**TEST_PHOTOMETRY_SETTINGS)
        settings_file.save(full_settings)
        assert settings_file.settings_file.exists()
        assert not settings_file.partial_settings_file.exists()
        assert settings_file.settings == full_settings
        assert settings_file.partial_settings is None

    @pytest.mark.parametrize("update", [True, False])
    def test_save_full_then_partial_settings(self, update):
        # Test that saving full settings and then partial settings generates
        # the expected error.
        settings_file = PhotometryWorkingDirSettings()
        full_settings = PhotometrySettings(**TEST_PHOTOMETRY_SETTINGS)
        settings_file.save(full_settings)
        assert settings_file.settings_file.exists()
        assert not settings_file.partial_settings_file.exists()
        assert settings_file.settings == full_settings

        camera = Camera.model_validate_json(CAMERA)
        # Change the camera name we we can detect whether the setting saved
        # to the working directory has been updated.
        camera.name = "new camera"
        partial_settings = PartialPhotometrySettings(camera=camera)

        if update:
            settings_file.save(partial_settings, update=update)
        else:
            error_message = (
                "Cannot save partial settings when full settings already exist"
            )
            with pytest.raises(ValueError, match=error_message):
                settings_file.save(partial_settings, update=update)

        assert not settings_file.partial_settings_file.exists()
        assert settings_file.settings_file.exists()
        assert settings_file.partial_settings is None
        if update:
            assert settings_file.settings.camera.name == camera.name
        else:
            assert settings_file.settings.camera.name == full_settings.camera.name

    def test_load_conflicting_partial_and_full_settings(self):
        # Make a valid partial settings file and a valid full settings file
        # that conflict with each other.
        settings_file = PhotometryWorkingDirSettings()
        camera = Camera.model_validate_json(CAMERA)
        partial_settings = PartialPhotometrySettings(camera=camera)
        # write these settings directly to the directory to avoid any conflict
        # resolution in the save method.
        with settings_file.partial_settings_file.open("w") as f:
            f.write(partial_settings.model_dump_json())

        full_settings = PhotometrySettings(**TEST_PHOTOMETRY_SETTINGS)
        with settings_file.settings_file.open("w") as f:
            f.write(full_settings.model_dump_json())

        # Try to load the settings. This should raise an error because the
        # settings conflict.
        error_message = "Partial settings and full settings do not match"
        with pytest.raises(ValueError, match=error_message):
            settings_file.load()

    def test_load_partial_and_full_both_valid(self):
        # Make a valid partial settings file that is actually a full file and
        # a valid full settings file that do not conflict with each other.
        settings_file = PhotometryWorkingDirSettings()

        partial_settings = PartialPhotometrySettings(**TEST_PHOTOMETRY_SETTINGS)
        # write these settings directly to the directory to avoid any conflict
        # resolution in the save method.
        with settings_file.partial_settings_file.open("w") as f:
            f.write(partial_settings.model_dump_json())

        full_settings = PhotometrySettings(**TEST_PHOTOMETRY_SETTINGS)
        with settings_file.settings_file.open("w") as f:
            f.write(full_settings.model_dump_json())

        # Load the settings. This should work because the settings don't conflict.
        settings = settings_file.load()
        assert settings == full_settings

        # Should have no partial settings
        assert not settings_file.partial_settings_file.exists()
        assert settings_file.partial_settings is None

    def test_load_no_settings(self):
        # Test that loading settings when no settings exist raise error.
        settings_file = PhotometryWorkingDirSettings()
        error_message = f"Settings file {settings_file.settings_file} does not exist"
        with pytest.raises(ValueError, match=error_message):
            settings_file.load()

    @pytest.mark.parametrize("full_settings", [True, False])
    def test_load_bad_settings(self, full_settings):
        # Test that loading bad settings raises an error.
        settings_file = PhotometryWorkingDirSettings()
        if full_settings:
            file = settings_file.settings_file
        else:
            file = settings_file.partial_settings_file
        with file.open("w") as f:
            f.write("{bad: settings}")
        error_message = "Error loading "
        with pytest.raises(ValueError, match=error_message):
            settings_file.load()

    def test_load_one_setting_present(self):
        # Test that loading settings when only one setting is present raises an error.
        settings_file = PhotometryWorkingDirSettings()
        partial_settings = PartialPhotometrySettings()
        with settings_file.partial_settings_file.open("w") as f:
            f.write(partial_settings.model_dump_json())

        settings = settings_file.load()
        assert settings == partial_settings

    def test_save_updates_instead_of_replacing(self):
        # Test that saving settings adds to whatever partial settings have
        # already been saved instead of dumping anything that used to be
        # there.
        settings_file = PhotometryWorkingDirSettings()

        # Save a Camera first
        partial_settings_cam = PartialPhotometrySettings(
            camera=TEST_PHOTOMETRY_SETTINGS["camera"]
        )
        settings_file.save(partial_settings_cam, update=True)
        from_file = settings_file.load()
        # Make sure the camera is there
        assert from_file.camera == Camera(**TEST_PHOTOMETRY_SETTINGS["camera"])

        # Save a different item, like an observatory
        partial_settings_obs = PartialPhotometrySettings(
            observatory=Observatory(**TEST_PHOTOMETRY_SETTINGS["observatory"])
        )
        settings_file.save(partial_settings_obs, update=True)
        from_file2 = settings_file.load()
        # Make sure the camera is still there
        assert from_file2.camera == Camera(**TEST_PHOTOMETRY_SETTINGS["camera"])
        # Make sure the observatory is there
        assert from_file2.observatory == Observatory(
            **TEST_PHOTOMETRY_SETTINGS["observatory"]
        )

    def test_save_update_completing_partial_makes_full(self):
        # Test that saving a partial settings file and then updating it to a
        # full settings file works.
        settings_file = PhotometryWorkingDirSettings()
        almost_complete_settings = TEST_PHOTOMETRY_SETTINGS.copy()
        the_observatory = almost_complete_settings.pop("observatory")

        # Make and save an object that has all settings except observatory
        partial_settings = PartialPhotometrySettings(**almost_complete_settings)
        settings_file.save(partial_settings, update=True)
        from_file = settings_file.load()
        # Make sure we have the partial settings
        assert from_file == partial_settings

        # Save the observatory
        the_last_setting = PartialPhotometrySettings(observatory=the_observatory)
        settings_file.save(the_last_setting, update=True)
        from_file2 = settings_file.load()
        # Make sure we have the full settings
        assert from_file2 == PhotometrySettings(**TEST_PHOTOMETRY_SETTINGS)
        # Make sure we have no partial settings
        assert settings_file.partial_settings is None
